/*
 * (C) 2003 Clemson University
 *
 * See COPYING in top-level directory.
 */

/*
 * This simple visualization watches a socket for the temperature values of a
 * region of steel plate.  It then displays them in a 3D world where the Z-axis
 * is a representation of how "hot" the plate it.  Very simple.
 */

#include <math.h>			
#include <time.h>
#include <stdio.h>		
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <values.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netdb.h>
#include <netinet/in.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <GL/gl.h>	
#include <GL/glu.h>	
#include <GL/glx.h>
#include "SDL.h"
#include "SDL_ttf.h"
#include "sockio.h"

#define BOOL    int
#define FALSE   0
#define TRUE    1

#define FONT_SIZE 18

#define MIN(a,b) (a < b) ? a : b

int MAKING_MOVIE = FALSE;
int movie_number = 0;

int SCWIDTH;
int SCHEIGHT;

typedef struct header {
	int numprocs;
	int MX;
	int MY;
	float HOT;
	float COLD;
	int iteration;
} header;

double *data;
#define DATA(i,j) data[i+(j*DATA_COLS)]
int DATA_COLS = -1;
int DATA_ROWS = -1;
float HOT;
float COLD;
int numprocs;

header info_header;
int numpoints = -1;
int iteration = -1;

/* socket stuff */
struct sockaddr_in mysin, fsin;

int two_d = FALSE;
int cur = 0;
int paused = 0;
float cube_size = 0.5;
int def_colors = 0;

int checkpoint_fd = -1;
int file_input = FALSE;

int connected = 0;

int drawAxes = TRUE;
int drawSticks = FALSE;
int draw2DWindow = TRUE;

GLUquadricObj *quad;

Uint8*		keys;					
BOOL		active=TRUE;		
BOOL		fullscreen=FALSE;	

GLfloat		xrot,yrot,zrot,
			xspeed,yspeed,zspeed,
			cx,cy,cz=-15;			

int			key=1;			
GLuint texture, base;

SDL_Surface *screen;
SDL_Rect dstrect;
SDL_Surface *text;
SDL_Color black = {0x00, 0x00, 0x00, 0};
SDL_Color font_color = {0xFF, 0x00, 0x00, 0};
int font_size = 16;
unsigned char *font_path = "./ganymede.ttf";

typedef struct				
{
	float	x, y, z;		
} VERTEX;			

void Quit(int retcode);

GLvoid ReSizeGLScene(GLsizei width, GLsizei height)	
{
	if (height==0)									
	{
		height=1;								
	}

	glViewport(0,0,width,height);					

	glMatrixMode(GL_PROJECTION);				
	glLoadIdentity();							

	if(two_d)
		glOrtho(0.0f, (GLfloat)width, height, 0.0f, -1.0f, 1.0f);
	else
		gluPerspective(45.0f,(GLfloat)width/(GLfloat)height,0.001f,1000.0f);

	glMatrixMode(GL_MODELVIEW);	
	glLoadIdentity();				
}

/* function to build our font list */
GLvoid buildFont( GLvoid )
{
	Display *dpy;          /* Our current X display */
	XFontStruct *fontInfo; /* Our font info */
	/* Sotrage for 96 characters */
	base = glGenLists( 96 );
	/* Get our current display long enough to get the fonts */
	dpy = XOpenDisplay( NULL );
	/* Get the font information */
	fontInfo = XLoadQueryFont( dpy, "9x15");
	/* If the above font didn't exist try one that should */
	if ( fontInfo == NULL ) {
		fontInfo = XLoadQueryFont( dpy, "fixed" );
		/* If that font doesn't exist, something is wrong */
		if ( fontInfo == NULL ) {
			fprintf( stderr, "no X font available?\n" );
			Quit( 1 );
		}
	}
	/* generate the list */
	glXUseXFont( fontInfo->fid, 32, 96, base );
	/* Recover some memory */
	XFreeFont( dpy, fontInfo );
	/* close the display now that we're done with it */
	XCloseDisplay( dpy );
	return;
}

/* Print our GL text to the screen */
GLvoid glPrint( const char *fmt, ... )
{
	char text[256]; /* Holds our string */
	va_list ap;     /* Pointer to our list of elements */

	/* If there's no text, do nothing */
	if ( fmt == NULL ) return;
	/* Parses The String For Variables */
	va_start( ap, fmt );
	/* Converts Symbols To Actual Numbers */
	vsprintf( text, fmt, ap );
	va_end( ap );
	/* Pushes the Display List Bits */
	glPushAttrib( GL_LIST_BIT );
	/* Sets base character to 32 */
	glListBase( base - 32 );
	/* Draws the text */ 
	glCallLists( strlen( text ), GL_UNSIGNED_BYTE, text );
	/* Pops the Display List Bits */
	glPopAttrib( );
}

/* function to recover memory form our list of characters */
GLvoid KillFont( GLvoid )
{
	glDeleteLists( base, 96 );
	return;
}

void WritePPM(char *filename, unsigned char *pixels, int xdim, int ydim)
{
	FILE *fp;

	fp = fopen(filename, "w");
	if(fp == NULL) {
		fprintf(stderr, "WritePPM() cannot create the PPM file: %s\n", filename);
		return;
	}

	fprintf(fp, "P6\n%d %d\n255\n", xdim, ydim);
	fwrite(pixels, sizeof(int), ydim*xdim, fp);

	fflush(fp);
	fclose(fp);
}

void drawSquare(float x, float y, float z, float th, float cr, float cg, float cb)
{
	glColor3f(cr, cg, cb);
	glBegin(GL_QUADS);
	glVertex3f(x-th,y-th,z);
	glVertex3f(x+th,y-th,z);
	glVertex3f(x+th,y+th,z);
	glVertex3f(x-th,y+th,z);
	glEnd();
}

void drawCube(float x, float y, float z1, float z2, 
	float th, float cr, float cg, float cb)
{
	glColor3f(cr, cg, cb);
	/* top */
	glBegin(GL_QUADS);

	glVertex3f(x-th,y-th,z1);
	glVertex3f(x+th,y-th,z1);
	glVertex3f(x+th,y-th,z2);
	glVertex3f(x-th,y-th,z2);

	glVertex3f(x-th,y+th,z1);
	glVertex3f(x+th,y+th,z1);
	glVertex3f(x+th,y+th,z2);
	glVertex3f(x-th,y+th,z2);

	glVertex3f(x+th,y-th,z1);
	glVertex3f(x+th,y+th,z1);
	glVertex3f(x+th,y+th,z2);
	glVertex3f(x+th,y-th,z2);

	glVertex3f(x-th,y-th,z1);
	glVertex3f(x-th,y+th,z1);
	glVertex3f(x-th,y+th,z2);
	glVertex3f(x-th,y-th,z2);

	glVertex3f(x-th,y-th,z1);
	glVertex3f(x+th,y-th,z1);
	glVertex3f(x+th,y+th,z1);
	glVertex3f(x-th,y+th,z1);

	glVertex3f(x-th,y-th,z2);
	glVertex3f(x+th,y-th,z2);
	glVertex3f(x+th,y+th,z2);
	glVertex3f(x-th,y+th,z2);
	glEnd();
}

void getColors(float data, float *r, float *g, float *b)
{
	float dist = HOT - COLD;
	float offset = data - COLD;

	*r = offset / dist;
	*g = 0.0;
	*b = 1.0 - (offset / dist);
}

void drawLegend(void)
{
	int leg_width = cube_size*(DATA_ROWS+1);
	int leg_height = cube_size*2;
	int xoff;
	int yoff;

	xoff = SCWIDTH/2 - leg_width/2;
	yoff = 0.8*SCHEIGHT;
	//yoff = (SCHEIGHT-(DATA_ROWS*cube_size)) + 0.1*SCHEIGHT;
	//xoff = (SCWIDTH-(DATA_COLS*cube_size)) + 0.3*SCWIDTH;
	//yoff = (SCHEIGHT-(DATA_ROWS*cube_size))/2.0 - cube_size;
	
	glColor3f(1.0, 0.0, 0.0);
	glBegin(GL_QUADS);
	glVertex2d(xoff, yoff);
	glVertex2d(xoff, yoff+leg_height);
	glColor3f(0.0, 0.0, 1.0);
	glVertex2d(xoff+leg_width, yoff+leg_height);
	glVertex2d(xoff+leg_width, yoff);
	glEnd();
	
	glColor3f(0.0f, 1.0f, 0.0f);
	glDisable(GL_DEPTH_TEST); 
	{
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		{
			glLoadIdentity();
			glMatrixMode(GL_MODELVIEW);
			glPushMatrix();
			{
				char str[128];
				glLoadIdentity();
 				gluOrtho2D(0, SCWIDTH, 0, SCHEIGHT);

				sprintf(str, "%3.0f", HOT);
				glRasterPos2f(xoff, SCHEIGHT-yoff);
				glPrint(str);

				sprintf(str, "%3.0f", COLD);
				glRasterPos2f(xoff+leg_width, SCHEIGHT-yoff);
				glPrint(str);
			}
			glPopMatrix();
		}
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
	}
	glMatrixMode(GL_MODELVIEW);
}

void Quit(int ret) 
{
	KillFont();
	SDL_Quit();
	exit(ret);
	//objfree(&bodies_obj);
}

int InitGL(GLvoid)			
{
	// Set The Blending Function For Translucency
	glBlendFunc(GL_SRC_ALPHA,GL_ONE);	
	// This Will Clear The Background Color To Black
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);		
	// Enables Clearing Of The Depth Buffer
	glClearDepth(1.0);	
	// The Type Of Depth Test To Do
	glDepthFunc(GL_LESS);	
	// Enables Depth Testing
	glDisable(GL_DEPTH_TEST); 
	//glEnable(GL_DEPTH_TEST);	
	//glEnable(GL_BLEND);
	// Enables Smooth Color Shading
	glShadeModel(GL_SMOOTH);	
	// Really Nice Perspective Calculations
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	quad = gluNewQuadric();
	gluQuadricNormals(quad, GLU_SMOOTH);
	gluQuadricTexture(quad, GL_TRUE);

	buildFont();

	return TRUE;								
}

void CheckSockets(int sock_connected)
{
	int retval;
	fd_set readfds;
	struct timeval now;

	if(connected) {
		FD_ZERO(&readfds);
		FD_SET(sock_connected, &readfds);
		now.tv_sec = 0;
		now.tv_usec = 0;
		retval = select(sock_connected+1, &readfds, NULL, NULL, &now);
		if(retval > 0) {
			if(FD_ISSET(sock_connected, &readfds)) {
				if(recv(sock_connected, &info_header, sizeof(header),
					MSG_WAITALL)!=sizeof(header)) {
				}
				else {
					if(DATA_COLS == -1) {
						DATA_COLS = info_header.MX;
						DATA_ROWS = info_header.MY;
						numpoints = DATA_COLS*DATA_ROWS;
						numprocs = info_header.numprocs;
						HOT = info_header.HOT;
						COLD = info_header.COLD;
						data = (double*)malloc(DATA_COLS*DATA_ROWS*sizeof(double));
						{
							int smaller = MIN(SCHEIGHT,SCWIDTH);
							int smaller2 = MIN(DATA_ROWS,DATA_COLS);

							cube_size = smaller/smaller2;
							cube_size /= 1.5;
						}
					}
					iteration = info_header.iteration;
					retval = brecv(sock_connected, data, DATA_COLS*DATA_ROWS*
						sizeof(double));
				}
			}
		}
		return;
	}
}

void draw2DGrid(void)
{
	int i;

	glDisable(GL_DEPTH_TEST);
	{
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		{
			glLoadIdentity();
			glMatrixMode(GL_MODELVIEW);
			glPushMatrix();
			{
				glLoadIdentity();
				glOrtho(0.0f, SCWIDTH, SCHEIGHT, 0.0f, -1.0f, 1.0f);
				if(numpoints > 0) {
					for(i=0; i<numpoints; i++)	
					{												
						float r, g, b;
						getColors(data[i], &r, &g, &b);
						drawSquare(
							(((int)(i/DATA_COLS))*cube_size*0.5),
							((i%DATA_COLS)*cube_size*0.5),
							0.0,
							cube_size*0.5, r, g, b);
					}
				}
			}
			glPopMatrix();
		} 
		glMatrixMode(GL_PROJECTION); 
		glPopMatrix();
	}
	glMatrixMode(GL_MODELVIEW);
}

void draw2DText(void)
{
	glColor3f(0.0f, 1.0f, 0.0f);
	glDisable(GL_DEPTH_TEST); 
	{
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		{
			glLoadIdentity();
			glMatrixMode(GL_MODELVIEW);
			glPushMatrix();
			{
				int numlines = 0;
				int xoffset = 5;
				int yoffset = 5;
				glLoadIdentity();
 				gluOrtho2D(0, SCWIDTH, 0, SCHEIGHT);
				//glRasterPos2f(xoffset, yoffset + FONT_SIZE*numlines++);
				//if(cubes) glPrint("Draw Mode: CUBES");
				//else glPrint("Draw Mode: POINTS");
				if(numpoints != -1 && iteration != -1) {
					char str[128];
					sprintf(str, "Iteration: %d", iteration);
					glRasterPos2f(xoffset, yoffset + FONT_SIZE*numlines++);
					glPrint(str);
					sprintf(str, "Grid: %d x %d (%d points)", DATA_COLS, DATA_ROWS,
						numpoints);
					glRasterPos2f(xoffset, yoffset + FONT_SIZE*numlines++);
					glPrint(str);
					sprintf(str, "Temperature: %3.1f (BLUE) -> %3.1f (RED) (degrees Kelvin)",
						COLD, HOT);
					glRasterPos2f(xoffset, yoffset + FONT_SIZE*numlines++);
					glPrint(str);
				}
				if(paused) {
					glRasterPos2f(xoffset, yoffset + FONT_SIZE*numlines++);
					glPrint("[ PAUSED ]");
				}
			}
			glPopMatrix();
		}
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
	}
	glMatrixMode(GL_MODELVIEW);
}

int DrawGLScene(int sock_accept, int *sock_connected)
{
	int i;
	fd_set readfds;
	struct timeval now;
	int retval, fromlen;

	if(file_input == FALSE) {
		FD_ZERO(&readfds);
		FD_SET(sock_accept, &readfds);
		now.tv_sec = 0;
		now.tv_usec = 0;
		retval = select(sock_accept+1, &readfds, NULL, NULL, &now);
		if(retval > 0) {
			if(FD_ISSET(sock_accept, &readfds)) {
				if((*sock_connected=accept(sock_accept,(struct sockaddr*)&fsin,
					&fromlen))!=-1) {
					printf("[Connection established]\n");
					if(connected) {
						/* if we were already connected, then we need to free some
						 * memory and restart */
						DATA_COLS = DATA_ROWS = iteration = numpoints = -1;
						free(data);
					}
					connected = 1;
				}
				else {
					perror("accept");
					exit(1);
				}
			}
		}
		if(connected && !paused) CheckSockets(*sock_connected);
	}

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();								

	if(!two_d) {
		glTranslatef(cx, cy, cz);
		glRotatef(xrot,1,0,0);
		glRotatef(yrot,0,1,0);
		glRotatef(zrot,0,0,1);
	
		xrot+=xspeed; yrot+=yspeed; zrot+=zspeed;
	
		glPushMatrix();
		// glTranslatef subtracts these numbers from all drawn below
		glTranslatef(-0.5, -0.5, 0.0);
	}

	if(numpoints > 0) {
		for(i=0; i<numpoints; i++)	
		{												
			float r, g, b;
			getColors(data[i], &r, &g, &b);
			if(two_d) {
				drawSquare(
					((((int)(i/DATA_COLS))*cube_size) + 
						((SCWIDTH/2 - (DATA_COLS*(cube_size))/2))),
					((((i%DATA_COLS)*cube_size) + 
						(SCHEIGHT-(DATA_ROWS*cube_size))/2.0)),
					0.0,
					cube_size, r, g, b);
			}
			else {
				drawSquare(
					((((int)(i/DATA_COLS))*cube_size) + 
						((SCWIDTH/2 - (DATA_COLS*(cube_size))/2)))/SCWIDTH,
					((((i%DATA_COLS)*cube_size) + 
						(SCHEIGHT-(DATA_ROWS*cube_size))/2.0))/SCHEIGHT,
					0.0,
					0.01, r, g, b
				);
				if(drawSticks) {
					glLineWidth(7.0f);
					//glColor3f(0.0, r, 0.0);
					glColor3f(0.0, r, 0.0);
					glBegin(GL_LINES);
						glVertex3d(
							((((int)(i/DATA_COLS))*cube_size) + 
								((SCWIDTH/2 - (DATA_COLS*(cube_size))/2)))/SCWIDTH,
							((((i%DATA_COLS)*cube_size) + 
								(SCHEIGHT-(DATA_ROWS*cube_size))/2.0))/SCHEIGHT,
							0.01
						);
						glVertex3d(
							((((int)(i/DATA_COLS))*cube_size) + 
								((SCWIDTH/2 - (DATA_COLS*(cube_size))/2)))/SCWIDTH,
							((((i%DATA_COLS)*cube_size) + 
								(SCHEIGHT-(DATA_ROWS*cube_size))/2.0))/SCHEIGHT,
							r+0.01
						);
					glEnd();
				}
				else {
					glColor3f(0.0, 1.0, 0.0);
					glBegin(GL_POINTS);
					glVertex3f(
						((((int)(i/DATA_COLS))*cube_size) + 
							((SCWIDTH/2 - (DATA_COLS*(cube_size))/2)))/SCWIDTH,
						((((i%DATA_COLS)*cube_size) + 
							(SCHEIGHT-(DATA_ROWS*cube_size))/2.0))/SCHEIGHT,
						r+0.01
					);
					glEnd();
				}
			}
		}
	}
	if(!two_d) glPopMatrix();

	if(!two_d && draw2DWindow) draw2DGrid();
	draw2DText();

	if(drawAxes && !two_d) {
		glLineWidth(1.5f);
		glColor3f(0.7f, 0.7f, 0.7f);
		glEnable(GL_LINE_STIPPLE);
		// Set the stippling pattern
		glLineStipple(3, 0xAAAA);
		glBegin(GL_LINES);
			glVertex3f(-0.3,0,0);
			glVertex3f(0.3,0,0);
			
			glVertex3f(0,-0.3,0);
			glVertex3f(0,0.3,0);
	
			glVertex3f(0,0,-0.3);
			glVertex3f(0,0,0.3);
		glEnd();
		glDisable(GL_LINE_STIPPLE);

		glColor3f(0.0f, 1.0f, 0.0f);
		glRasterPos3f(-0.3,0,0);
		glPrint("-X");
		glRasterPos3f(0.3,0,0);
		glPrint("+X");
		glRasterPos3f(0,-0.3,0);
		glPrint("-Y");
		glRasterPos3f(0,0.3,0);
		glPrint("+Y");
		glRasterPos3f(0,0,-0.3);
		glPrint("-Z");
		glRasterPos3f(0,0,0.3);
		glPrint("+Z");
	}

	if(MAKING_MOVIE && ((iteration != movie_number) || paused)) {
		unsigned char *pixels, *fn;

		pixels = (unsigned char*)malloc(3*SCWIDTH*SCHEIGHT*sizeof(unsigned char));
		fn = (unsigned char*)malloc(80);
		
		glReadBuffer(GL_FRONT);
		glReadPixels(0,0,SCWIDTH,SCHEIGHT,GL_RGB,GL_UNSIGNED_BYTE,pixels);

		sprintf(fn, "/tmp/sshot-%05d.ppm", movie_number);
		WritePPM(fn, pixels, SCWIDTH, SCHEIGHT);

		free(pixels);
		free(fn);

		movie_number++;
	} 

	return TRUE;
}


/*	This Code Creates Our OpenGL Window.  Parameters Are:					*
 *	title			- Title To Appear At The Top Of The Window				*
 *	width			- Width Of The GL Window Or Fullscreen Mode				*
 *	height			- Height Of The GL Window Or Fullscreen Mode			*
 *	bits			- Number Of Bits To Use For Color (8/16/24/32)			*
 *	fullscreenflag	- Use Fullscreen Mode (TRUE) Or Windowed Mode (FALSE)	*/
BOOL CreateGLWindow(char* title, int width, int height, int bits, 
	BOOL fullscreenflag)
{
	Uint32 flags;

	fullscreen=fullscreenflag;			// Set The Global Fullscreen Flag
	flags = SDL_OPENGL;
	if ( fullscreenflag ) {
		flags |= SDL_FULLSCREEN;
	}
	if ( (screen = SDL_SetVideoMode(width, height, 0, flags)) == NULL ) {
		return FALSE;
	}
	SDL_WM_SetCaption(title, "glpv: OpenGL Particle Visualization");

	// Set Up Our Perspective GL Screen
	ReSizeGLScene(width, height);							

	// Initialize Our Newly Created GL Window
	if (!InitGL())									
	{
		Quit(1);	// Reset The Display
		return FALSE;			// Return FALSE
	}

	return TRUE;				// Success
}

int TOGGLE(int x) {
	if(x == 1) return 0;
	else return 1;
}

int main(int argc, char *argv[])
{
	BOOL	done=FALSE;
	int sock_accept = -1, sock_connected;
	float zoomz = 0.2;
	
	if(argc < 4) {
		fprintf(stderr,"Syntax: %s <screen width> <screen height> <port to bind to OR checkpoint file to read from>\n", argv[0]);
		exit(1);
	}
	SCWIDTH = atoi(argv[1]);
	SCHEIGHT = atoi(argv[2]);

	srand((unsigned int)time(NULL));

	if(atoi(argv[3]) == 0) {
		checkpoint_fd = open(argv[3], O_RDONLY);
		if(checkpoint_fd == -1) {
			perror("open");
			exit(1);
		}
		file_input = TRUE;
	}
	
	else {
		sock_accept = new_sock();
		bind_sock(sock_accept, atoi(argv[3]));

		if(listen(sock_accept,5)<0) {
			perror("listen");
			exit(1);
		}
	}

	/* Initialize SDL */
	if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
		fprintf(stderr, "Couldn't init SDL: %s\n", SDL_GetError());
		return 1;
	}

	// Create Our OpenGL Window
	if (!CreateGLWindow("3D Body Visualization",SCWIDTH,SCHEIGHT,16,fullscreen))
	{
		SDL_Quit();
		return 0;									
	}

	/* initial view of the world */
	xrot = -77.5;
	yrot = 4.5;
	zrot = -36.5;
	cx = 0.0;
	cy = -0.1;
	cz = -1.019920;

	printf("COMMANDS:\n");
	printf("F1 - toggle full-screen mode\n");
	printf("ESC - quit\n");
	printf("p - TOGGLE pause\n");
	printf("t - TOGGLE drawing 2D or 3D\n");
	printf("m - TOGGLE movie making (look for screenshots in /tmp)\n");
	printf("3D Commands:\n");
	printf("\tPGUP/PGDWN - rotate Z axis\n");
	printf("\tKEYPAD LEFT/RIGHT - rotate X axis\n");
	printf("\tKEYPAD UP/DOWN - rotate Y axis\n");
	printf("\tLEFT SHIFT + any of the above rotations - put spin on that axis\n");
	printf("\tw/s - zoom in / out\n");
	printf("\ta/d - move view field left / right\n");
	printf("\tq/z - move field view up / down\n");
	printf("\tv - TOGGLE drawing axes\n");
	printf("\tg - TOGGLE drawing temperature sticks or points\n");
	printf("\tr - TOGGLE drawing 2D region in top left window\n");
	printf("\tk - return or original view\n");
	printf("\tx - stop all axial spins\n");

	while(!done)								
	{
		SDL_Event event;
		while ( SDL_PollEvent(&event) ) {
			switch (event.type) {
				case SDL_QUIT:
					done=TRUE;				
					break;
				case SDL_KEYDOWN:
					switch(event.key.keysym.sym) {
						case SDLK_m:
							MAKING_MOVIE = MAKING_MOVIE ? 0 : 1;
							if(MAKING_MOVIE)
								printf("Now making movie, look for screenshots in /tmp\n");
							else
								printf("Movie making turned off.\n");
							break;
						case SDLK_r:
							if(!two_d)
								draw2DWindow = draw2DWindow ? 0 : 1;
							break;
						case SDLK_g:
							if(!two_d)
								drawSticks = drawSticks ? 0 : 1;
							break;
						case SDLK_v:
							if(!two_d)
								drawAxes = drawAxes ? 0 : 1;
							break;
						case SDLK_t:
							two_d = two_d ? 0 : 1;
							ReSizeGLScene(SCWIDTH, SCHEIGHT);							
							break;
						case SDLK_p:
							paused = paused ? 0 : 1;
							break;
						case SDLK_k:
							if(!two_d) {
								xrot = -77.5;
								yrot = 4.5;
								zrot = -36.5;
								cx = 0.0;
								cy = -0.1;
								cz = -1.019920;
							}
							break;
						default: break;
					}
				default:
					break;
			}
		}
		keys = SDL_GetKeyState(NULL);
		// Draw The Scene.  Watch For ESC Key And Quit Messages From DrawGLScene()
		if ((active && !DrawGLScene(sock_accept, &sock_connected)) || keys[SDLK_ESCAPE])
		{
			done=TRUE;					
		}
		else							
		{
			SDL_GL_SwapBuffers();					// Swap Buffers (Double Buffering)
		}

		if(!two_d) {
			if(keys[SDLK_PAGEUP]) {
				if(keys[SDLK_LSHIFT]) zspeed += 0.01f;
				else zrot += 0.5f;
			}
			if(keys[SDLK_PAGEDOWN]) {
				if(keys[SDLK_LSHIFT]) zspeed -= 0.01f;
				else zrot -= 0.5f;
			}
			if(keys[SDLK_DOWN]) {
				if(keys[SDLK_LSHIFT]) xspeed += 0.01f;
				else xrot += 0.5f;
			}
			if(keys[SDLK_UP]) {
				if(keys[SDLK_LSHIFT]) xspeed -= 0.01f;
				else xrot -= 0.5f;
			}
			if(keys[SDLK_RIGHT]) {
				if(keys[SDLK_LSHIFT]) yspeed += 0.01f;
				else yrot += 0.5f;
			}
			if(keys[SDLK_LEFT]) {
				if(keys[SDLK_LSHIFT]) yspeed -= 0.01f;
				else yrot -= 0.5f;
			}
	
			/* stop the spinning */
			if(keys[SDLK_x]) xspeed = yspeed = zspeed = 0.0f;
	
			// Move Object Away From Viewer
			if (keys[SDLK_s])	{
				cz -= zoomz;
			}
	
			// Move Object Towards Viewer
			if (keys[SDLK_w])	{
				cz+=zoomz;
			}
	
			// Move Object Up
			if (keys[SDLK_z])	cy+=0.1f;				
	
			// Move Object Down
			if (keys[SDLK_q])	cy-=0.1f;				
	
			// Move Object Right
			if (keys[SDLK_a])	cx+=0.1f;				
	
			// Move Object Left
			if (keys[SDLK_d])	cx-=0.1f;										
		}

		if (keys[SDLK_F1])							
		{
			fullscreen=!fullscreen;		// Toggle Fullscreen / Windowed Mode
			// Recreate Our OpenGL Window
			if (!CreateGLWindow("3D CFD Visualization",SCWIDTH,SCHEIGHT,16,fullscreen))
			{
				done=TRUE;					// Quit If Window Was Not Created
			}
		}
	}

	// Shutdown
	close(sock_accept);
	Quit(0); // Kill The Window
	SDL_Quit();
	return 0;							
}
